--- ufraw_preview.c.orig	Mon May  7 10:43:23 2007
+++ ufraw_preview.c	Mon May  7 18:20:11 2007
@@ -43,8 +43,12 @@
 const int def_preview_height = 9000;
 #define raw_his_size 320
 #define live_his_size 256
-#define min_scale 2
-#define max_scale 20
+#define min_zoom 10.
+#ifdef HAVE_GTKIMAGEVIEW
+#define max_zoom 200.
+#else
+#define max_zoom 50.
+#endif
 
 enum { pixel_format, percent_format };
 
@@ -488,25 +492,37 @@
 	int x, int y, int width, int height)
 {
     /* Find location of area in the view. */
+    int x1, y1;
+    gdouble zoom = gtk_image_view_get_zoom(view);
     GdkRectangle viewRect;
     gtk_image_view_get_viewport(view, &viewRect);
+/*#ifdef DEBUG
+    fprintf(stderr,"image_view_draw_area: zoom: %lf; viewport: %dx%d%+d%+d\n", 
+	zoom,viewRect.width,viewRect.height,viewRect.x,viewRect.y);
+    fflush(stderr);
+#endif*/
     GdkPixbuf *pixbuf = gtk_image_view_get_pixbuf(view);
-    int pixbufWidth = gdk_pixbuf_get_width(pixbuf);
-    int pixbufHeight = gdk_pixbuf_get_height(pixbuf);
-    int viewWidth = GTK_WIDGET(view)->allocation.width;
-    int viewHeight = GTK_WIDGET(view)->allocation.height;
-    if ( pixbufWidth>viewWidth ) {
-	if (x<viewRect.x) width += x - viewRect.x;
-	x = MAX(x-viewRect.x, 0);
-    } else {
-	x += (viewWidth - pixbufWidth) / 2;
-    }
-    if ( pixbufHeight>viewHeight ) {
-	if (y<viewRect.y) height += y - viewRect.y;
-	y = MAX(y-viewRect.y, 0);
-    } else {
-	y += (viewHeight - pixbufHeight) / 2;
-    }
+    int pixbufWidth = (int)ceil(gdk_pixbuf_get_width(pixbuf) * zoom);
+    int pixbufHeight = (int)ceil(gdk_pixbuf_get_height(pixbuf) * zoom);
+    x = (int)floor(x * zoom);
+    y = (int)floor(y * zoom);
+    width = (int)ceil(width * zoom);
+    height = (int)ceil(height * zoom);
+    x1 = x + width;
+    y1 = y + height;
+    if (x1 > pixbufWidth) x1 = pixbufWidth;
+    if (y1 > pixbufHeight) y1 = pixbufHeight;
+    x = MAX(x, viewRect.x);
+    y = MAX(y, viewRect.y);
+    if (x > pixbufWidth) x = pixbufWidth;
+    if (y > pixbufHeight) y = pixbufHeight;
+    width = MIN(x1,viewRect.x + viewRect.width) - x;
+    height = MIN(y1,viewRect.y + viewRect.height) - y;
+/*#ifdef DEBUG
+    fprintf(stderr,"image_view_draw_area: draw area: %dx%d%+d%+d\n", 
+	width, height, x, y);
+    fflush(stderr);
+#endif*/
     if ( height>0 && width>0 )
 	gtk_widget_queue_draw_area(GTK_WIDGET(view), x, y, width, height);
 }
@@ -544,7 +560,7 @@
     static guint8 *pixies;
     int rowstride;
     guint8 pix[99], *p8, pen[4][3];
-    guint16 pixtmp[9999];
+    guint16 pixtmp[3];
     image_type p16;
     int x, c, cl, colors, y, y0, y1;
     int raw_his[raw_his_size][4], raw_his_max;
@@ -677,7 +693,7 @@
     static int width, height, x, y, c, o, min, max, rowstride, y0;
     static int live_his[live_his_size][4], live_his_max;
     guint8 *p8;
-    guint16 pixtmp[9999];
+    guint16 *pixtmp;
     double rgb[3];
     guint64 sum[3], sqr[3];
 
@@ -688,17 +704,18 @@
 	rowstride = gdk_pixbuf_get_rowstride(data->PreviewPixbuf);
 	pixies = gdk_pixbuf_get_pixels(data->PreviewPixbuf);
 #ifdef DEBUG
-	fprintf(stderr, "render_preview_image: w=%d, h=%d, r=%d, mode=%d\n",
+	fprintf(stderr, "render_preview_image: w=%d, h=%d, r=%d, mode=%ld\n",
 		width, height, rowstride, data->RenderMode);
 	fflush(stderr);
 #endif
 	memset(live_his, 0, sizeof(live_his));
 	y = y0 = 0;
     }
+    pixtmp = (guint16 *)alloca(2*rowstride);
     for (; y<height; y++) {
-        develope(&pixies[y*rowstride], data->UF->image.image[y*width],
-                Developer, 8, pixtmp, width);
-        for (x=0, p8=&pixies[y*rowstride]; x<width; x++, p8+=3) {
+        p8 = pixies + y * rowstride;
+        develope(p8, data->UF->image.image[y*width], Developer, 8, pixtmp, width);
+        for (x=0; x<width; x++, p8+=3) {
             for (c=0, max=0, min=0x100; c<3; c++) {
                 max = MAX(max, p8[c]);
                 min = MIN(min, p8[c]);
@@ -726,17 +743,16 @@
             }
         }
         if (y%32==31) {
+            int h = ++y - y0;
 #ifdef HAVE_GTKIMAGEVIEW
 //	    With the following line redraw does not work:
 //	    gtk_widget_queue_draw(data->PreviewWidget);
 	    image_view_draw_area(GTK_IMAGE_VIEW(data->PreviewWidget),
-		    0, y0, width, y+1-y0);
+		    0, y0, width, h);
 #else
-	    gtk_widget_queue_draw_area(data->PreviewWidget,
-		    0, y0, width, y+1-y0);
+	    gtk_widget_queue_draw_area(data->PreviewWidget, 0, y0, width, h);
 #endif
-            y0 = y+1;
-            y++;
+            y0 = y;
             return TRUE;
         }
     }
@@ -744,10 +760,10 @@
 //    gtk_image_view_set_pixbuf(GTK_IMAGE_VIEW(data->PreviewWidget),
 //	    data->PreviewPixbuf, FALSE);
     image_view_draw_area(GTK_IMAGE_VIEW(data->PreviewWidget),
-	    0, y0, width, y+1-y0);
+	    0, y0, width, height-y0);
     g_signal_emit_by_name(G_OBJECT(data->PreviewWidget), "pixbuf-changed");
 #else
-    gtk_widget_queue_draw_area(data->PreviewWidget, 0, y0, width, y+1-y0);
+    gtk_widget_queue_draw_area(data->PreviewWidget, 0, y0, width, height-y0);
 #endif
     /* draw live histogram */
     pixbuf = gtk_image_get_pixbuf(GTK_IMAGE(data->LiveHisto));
@@ -1147,16 +1163,24 @@
 
 void create_base_image(preview_data *data)
 {
+#ifdef HAVE_GTKIMAGEVIEW
+    if (data->UF->image.image == NULL)
+    {
+	ufraw_convert_image(data->UF);
+	data->PreviewPixbuf = gdk_pixbuf_new(GDK_COLORSPACE_RGB, FALSE, 8,
+		data->UF->image.width, data->UF->image.height);
+	gtk_image_view_set_pixbuf(GTK_IMAGE_VIEW(data->PreviewWidget), 
+		data->PreviewPixbuf, FALSE);
+	g_object_unref(data->PreviewPixbuf);
+    }
+    gtk_image_view_set_zoom(GTK_IMAGE_VIEW(data->PreviewWidget), 
+	CFG->Zoom / 100.);
+#else
     int shrinkSave = CFG->shrink;
     int sizeSave = CFG->size;
-    if (CFG->Scale==0) {
-	CFG->size = CFG->Zoom / 100.0 *
-	    MAX(data->UF->predictedHeight, data->UF->predictedWidth);
-	CFG->shrink = 0;
-    } else {
-	CFG->size = 0;
-	CFG->shrink = CFG->Scale;
-    }
+    CFG->size = CFG->Zoom / 100.0 *
+	MAX(data->UF->predictedHeight, data->UF->predictedWidth);
+    CFG->shrink = 0;
     g_free(data->UF->image.image);
     data->UF->image.image = NULL;
     ufraw_convert_image(data->UF);
@@ -1167,26 +1191,16 @@
     if (width!=data->UF->image.width || height!=data->UF->image.height) {
 	data->PreviewPixbuf = gdk_pixbuf_new(GDK_COLORSPACE_RGB, FALSE, 8,
 		data->UF->image.width, data->UF->image.height);
-#ifdef HAVE_GTKIMAGEVIEW
-	gtk_image_view_set_pixbuf(GTK_IMAGE_VIEW(data->PreviewWidget),
-		data->PreviewPixbuf, FALSE);
-	gtk_image_view_set_zoom(GTK_IMAGE_VIEW(data->PreviewWidget), 1.0);
-#else
 	gtk_image_set_from_pixbuf(GTK_IMAGE(data->PreviewWidget),
 		data->PreviewPixbuf);
-#endif
 	g_object_unref(data->PreviewPixbuf);
     }
-    char progressText[max_name];
-    if (CFG->Scale==0)
+#endif
+    if (data->ProgressBar!=NULL) {
+	char progressText[max_name];
 	snprintf(progressText, max_name, _("size %dx%d, zoom %2.f%%"),
 		data->UF->predictedWidth, data->UF->predictedHeight,
 		CFG->Zoom);
-    else
-	snprintf(progressText, max_name, _("size %dx%d, scale 1/%d"),
-		data->UF->predictedWidth, data->UF->predictedHeight,
-		CFG->Scale);
-    if (data->ProgressBar!=NULL) {
 	gtk_progress_bar_set_text(data->ProgressBar, progressText);
 	gtk_progress_bar_set_fraction(data->ProgressBar, 0);
     }
@@ -1197,16 +1211,8 @@
     preview_data *data = get_preview_data(widget);
     user_data = user_data;
     if (data->FreezeDialog) return;
-    if (CFG->Scale==0) {
-	if (CFG->Zoom<100.0/max_scale) CFG->Zoom = 100.0/max_scale;
-	CFG->Scale = floor(100.0/CFG->Zoom);
-	if (CFG->Scale==100.0/CFG->Zoom) CFG->Scale--;
-    } else {
-	CFG->Scale--;
-    }
-    if (CFG->Scale<min_scale) CFG->Scale = min_scale;
-    if (CFG->Scale>max_scale) CFG->Scale = max_scale;
-    CFG->Zoom = 100.0/CFG->Scale;
+    CFG->Zoom = (floor(CFG->Zoom/20.)+1.)*20.;
+    if (CFG->Zoom > max_zoom) CFG->Zoom = max_zoom;
     create_base_image(data);
     gtk_adjustment_set_value(data->ZoomAdjustment, CFG->Zoom);
     render_preview(data, render_default);
@@ -1217,16 +1223,8 @@
     preview_data *data = get_preview_data(widget);
     user_data = user_data;
     if (data->FreezeDialog) return;
-    if (CFG->Scale==0) {
-	if (CFG->Zoom<100.0/max_scale) CFG->Zoom = 100.0/max_scale;
-	CFG->Scale = ceil(100.0/CFG->Zoom);
-	if (CFG->Scale==100.0/CFG->Zoom) CFG->Scale++;
-    } else {
-	CFG->Scale++;
-    }
-    if (CFG->Scale<min_scale) CFG->Scale = min_scale;
-    if (CFG->Scale>max_scale) CFG->Scale = max_scale;
-    CFG->Zoom = 100.0/CFG->Scale;
+    CFG->Zoom = (floor(CFG->Zoom/20.)-1.)*20.;
+    if (CFG->Zoom < min_zoom) CFG->Zoom = min_zoom;
     create_base_image(data);
     gtk_adjustment_set_value(data->ZoomAdjustment, CFG->Zoom);
     render_preview(data, render_default);
@@ -1244,6 +1242,10 @@
 	data->UF->predictedWidth = data->UF->predictedHeight;
 	data->UF->predictedHeight = temp;
     }
+#ifdef HAVE_GTKIMAGEVIEW
+    free(data->UF->image.image);
+    data->UF->image.image = NULL;
+#endif
     create_base_image(data);
     render_preview(data, render_default);
 }
@@ -1545,9 +1547,12 @@
 	CFG->autoExposure = FALSE;
     }
     if (valuep==&CFG->Zoom) {
-	CFG->Scale = 0;
 	create_base_image(data);
     } else if (valuep==&CFG->threshold) {
+/*#ifdef HAVE_GTKIMAGEVIEW
+	free(data->UF->image.image);
+	data->UF->image.image = NULL;
+#endif*/
 	create_base_image(data);
     } else {
 	if (CFG->autoExposure==enabled_state) CFG->autoExposure = apply_state;
@@ -2140,14 +2145,17 @@
     uf->widget = previewWindow;
 
     gdk_screen_get_monitor_geometry(gdk_screen_get_default(), 0, &screen);
-    max_preview_width = MIN(def_preview_width, screen.width-416);
+    max_preview_width = MIN(def_preview_width, screen.width-432);
     max_preview_height = MIN(def_preview_height, screen.height-120);
-    CFG->Scale = MAX((uf->predictedWidth-1)/max_preview_width,
-            (uf->predictedHeight-1)/max_preview_height)+1;
-    CFG->Scale = MAX(2, CFG->Scale);
-    CFG->Zoom = 100.0 / CFG->Scale;
-    preview_width = uf->predictedWidth / CFG->Scale;
-    preview_height = uf->predictedHeight / CFG->Scale;
+    CFG->Zoom = 
+        floor(MIN(
+            (double)max_preview_width/uf->predictedWidth, 
+            (double)max_preview_height/uf->predictedHeight
+        ) * 100.);
+    if (CFG->Zoom < min_zoom) CFG->Zoom = min_zoom;
+    if (CFG->Zoom > max_zoom) CFG->Zoom = max_zoom;
+    preview_width = (int)floor(uf->predictedWidth * CFG->Zoom / 100.);
+    preview_height = (int)floor(uf->predictedHeight * CFG->Zoom / 100.);
     /* With the following guesses the window usually fits into the screen.
      * There should be more intelegent settings to widget sizes. */
     curveeditorHeight = 256;
@@ -2740,7 +2748,7 @@
     label = gtk_label_new(_("Zoom %"));
     gtk_table_attach(table, label, 0, 1, 0, 1, 0, 0, 0, 0);
     data->ZoomAdjustment = GTK_ADJUSTMENT(gtk_adjustment_new(
-		CFG->Zoom, 5, 50, 1, 1, 0));
+		CFG->Zoom, min_zoom, max_zoom, 1, 1, 0));
     g_object_set_data(G_OBJECT(data->ZoomAdjustment),
 		"Adjustment-Accuracy", (gpointer)0);
     button = gtk_spin_button_new(data->ZoomAdjustment, 1, 0);
@@ -3021,7 +3029,7 @@
     data->PreviewWidget = gtk_image_view_new();
     gtk_image_view_set_pixbuf(GTK_IMAGE_VIEW(data->PreviewWidget),
 	    data->PreviewPixbuf, FALSE);
-    gtk_image_view_set_zoom(GTK_IMAGE_VIEW(data->PreviewWidget), 1.0);
+    gtk_image_view_set_zoom(GTK_IMAGE_VIEW(data->PreviewWidget), CFG->Zoom / 100);
     gtk_widget_set_size_request(data->PreviewWidget,
 	    preview_width, preview_height);
     gtk_event_box_set_above_child(GTK_EVENT_BOX(PreviewEventBox), TRUE);
