--- dcraw.cc.orig	2016-08-06 07:02:57 UTC
+++ dcraw.cc
@@ -4527,6 +4527,7 @@ void CLASS wavelet_denoise()
 
   while (maximum << scale < 0x10000) scale++;
   maximum <<= --scale;
+  FORC4 max4[c] <<= scale;
   black <<= scale;
   FORC4 cblack[c] <<= scale;
   if ((size = iheight*iwidth) < 0x15550000)
@@ -4601,6 +4602,7 @@ void CLASS scale_colors()
   float scale_mul[4], fr, fc;
   ushort *img=0, *pix;
 
+  FORC4 if (max4[c] <= 0) max4[c] = maximum;
   if (user_mul[0])
     memcpy (pre_mul, user_mul, sizeof pre_mul);
   if (use_auto_wb || (use_camera_wb && cam_mul[0] == -1)) {
@@ -4618,7 +4620,7 @@ void CLASS scale_colors()
 		val = BAYER2(y,x);
 	      } else
 		val = image[y*width+x][c];
-	      if (val > (int)(maximum-25)) goto skip_block;
+	      if (val > (int)(max4[c]-25)) goto skip_block;
 	      if ((val -= cblack[c]) < 0) val = 0;
 	      sum[c] += val;
 	      sum[c+4]++;
@@ -4651,6 +4653,7 @@ skip_block: ;
   sat = maximum;
   if (threshold) wavelet_denoise();
   maximum -= black;
+  FORC4 max4[c] -= cblack[c];
   for (dmin=DBL_MAX, dmax=c=0; c < 4; c++) {
     if (dmin > pre_mul[c])
 	dmin = pre_mul[c];
@@ -4658,16 +4661,22 @@ skip_block: ;
 	dmax = pre_mul[c];
   }
   if (!highlight) dmax = dmin;
-  FORC4 scale_mul[c] = (pre_mul[c] /= dmax) * 65535.0 / maximum;
+  FORC4 scale_mul[c] = (pre_mul[c] /= dmax) * 65535.0 / max4[c];
   dcraw_message(DCRAW_VERBOSE,
       _("Scaling with darkness %d, saturation %d, and\nmultipliers"), dark, sat);
   FORC4 dcraw_message(DCRAW_VERBOSE, " %f", pre_mul[c]);
-  dcraw_message(DCRAW_VERBOSE, "\n");
   if (filters > 1000 && (cblack[4]+1)/2 == 1 && (cblack[5]+1)/2 == 1) {
     FORC4 cblack[FC(c/2,c%2)] +=
 	cblack[6 + c/2 % cblack[4] * cblack[5] + c%2 % cblack[5]];
     cblack[4] = cblack[5] = 0;
   }
+  dcraw_message(DCRAW_VERBOSE, _("\ncblacks: "));
+  FORC4 dcraw_message(DCRAW_VERBOSE, " %d", cblack[c]);
+  dcraw_message(DCRAW_VERBOSE, _("\nmaximum: %d; {"), maximum);
+  FORC4 dcraw_message(DCRAW_VERBOSE, " %d", max4[c]);
+  dcraw_message(DCRAW_VERBOSE, _(" }\nscaled multipliers: "));
+  FORC4 dcraw_message(DCRAW_VERBOSE, " %f", scale_mul[c]);
+  dcraw_message(DCRAW_VERBOSE, _("\nfilters: %08x\n"), filters);
   size = iheight*iwidth;
   for (i=0; i < size*4; i++) {
     if (!(val = ((ushort *)image)[i])) continue;
@@ -5387,8 +5396,8 @@ void CLASS ahd_interpolate()
 void CLASS median_filter()
 {
   ushort (*pix)[4];
-  unsigned c, i, j, k;
-  int pass, med[9];
+  unsigned c, k;
+  int i, j, pass, med[9];
   static const uchar opt[] =	/* Optimal 9-element median search */
   { 1,2, 4,5, 7,8, 0,1, 3,4, 6,7, 1,2, 4,5, 7,8,
     0,3, 5,8, 4,7, 3,6, 1,4, 2,5, 4,7, 4,2, 6,4, 4,2 };
@@ -5403,9 +5412,9 @@ void CLASS median_filter()
 	for (k=0, i = -width; i <= width; i += width)
 	  for (j = i-1; j <= i+1; j++)
 	    med[k++] = pix[j][3] - pix[j][1];
-	for (i=0; i < sizeof opt; i+=2)
-	  if     (med[opt[i]] > med[opt[i+1]])
-	    SWAP (med[opt[i]] , med[opt[i+1]]);
+	for (k=0; k < sizeof opt; k+=2)
+	  if     (med[opt[k]] > med[opt[k+1]])
+	    SWAP (med[opt[k]] , med[opt[k+1]]);
 	pix[0][c] = CLIP(med[4] + pix[0][1]);
       }
     }
@@ -7528,7 +7537,7 @@ void CLASS adobe_coeff (const char *make
 	{ 6257,-303,-1000,-7880,15621,2396,-1714,1904,7046 } },
     { "Canon EOS 40D", 0, 0x3f60,
 	{ 6071,-747,-856,-7653,15365,2441,-2025,2553,7315 } },
-    { "Canon EOS 50D", 0, 0x3d93,
+    { "Canon EOS 50D", 0, 0x3c34, // 0x3d93 is too optimistic
 	{ 4920,616,-593,-6493,13964,2784,-1774,3178,7005 } },
     { "Canon EOS 60D", 0, 0x2ff7,
 	{ 6719,-994,-925,-4408,12426,2211,-887,2129,6051 } },
@@ -8614,6 +8623,111 @@ float CLASS find_green (int bps, int bit
 }
 
 /*
+   Read RAW* environment variables and store values into statics.
+ */
+static int env_black=-1;
+static int env_sat[4] = { -1, -1, -1, -1 };
+void CLASS read_env()
+{
+  char const *e = getenv("RAWMIN");
+  if (e && *e) {
+    errno = 0;
+    int v = (int)strtol(e, (char **)NULL, 10);
+    if (errno) {
+      dcraw_message(DCRAW_WARNING, "cannot convert RAWMIN environment variable value (%s) to integer (err: %s; %d)\n", e, strerror(errno), errno);
+    } else {
+      env_black = v;
+    }
+  }
+  e = getenv("RAWMAX");
+  if (e && *e) {
+    //errno = 0;
+    //int v = (int)strtol(e, (char **)NULL, 10);
+    int v = sscanf(e, "%d%d%d%d", &env_sat[0], &env_sat[1], &env_sat[2], &env_sat[3]);
+    if (1 == v) {
+      env_sat[3] = env_sat[2] = env_sat[1] = env_sat[0];
+    } else if (4 != v) {
+      dcraw_message(DCRAW_WARNING, "cannot convert RAWMAX environment variable value (%s) to integer(s)\n", e);
+    } 
+  }
+}
+
+/*
+   Apply black & saturation values getting from environment and/or command line.
+ */
+void CLASS apply_black_sat(int const user_black, int const user_sat[])
+{
+  int c, i;
+  FORC4 max4[c] = 0;
+  // Canon EOS 50D has another black & maximum for some ISO values
+  if (!strcmp(make,"Canon") && !strcmp(model,"EOS 50D")) {
+    if (iso_speed<160) {
+      max4[0] = 0x3385;
+      max4[1] = 0x3477;
+      max4[3] = 0x3477;
+      max4[2] = 0x3353;
+      maximum = 0x3353;
+    } else if (iso_speed==160 || iso_speed==320 || iso_speed==640) {
+      //black = 0x0006;
+      max4[0] = 0x3123;
+      max4[1] = 0x3207;
+      max4[3] = 0x3207;
+      max4[2] = 0x30f3;
+      maximum = 0x30f3;
+    } else if (iso_speed==800 || iso_speed==1000) {
+      //black = 0x000b;
+      max4[0] = 0x3c6f;
+      max4[1] = 0x3d89;
+      max4[3] = 0x3d89;
+      max4[2] = 0x3c31;
+      maximum = 0x3c31;
+    } else if (iso_speed==1250) {
+      //black = 0x000b;
+      max4[0] = 0x3125;
+      max4[1] = 0x3203;
+      max4[3] = 0x3203;
+      max4[2] = 0x30ef;
+      maximum = 0x30ef;
+    } else if (iso_speed==1600 || iso_speed==2000 || iso_speed==2500) {
+      //black = 0x0012;
+      max4[0] = 0x3c6d;
+      max4[1] = 0x3d7f;
+      max4[3] = 0x3d7f;
+      max4[2] = 0x3c27;
+      maximum = 0x3c27;
+    } else {
+      max4[0] = 0x3c6f;
+      max4[1] = 0x3d8d;
+      max4[3] = 0x3d8d;
+      max4[2] = 0x3c35;
+    }
+  }
+  i = cblack[3];
+  FORC3 if (i > (int) cblack[c]) i = cblack[c];
+  FORC4 cblack[c] -= i;
+  black += i;
+  i = cblack[6];
+  FORC ((unsigned)(cblack[4] * cblack[5]))
+    if (i > cblack[6+c]) i = cblack[6+c];
+  FORC ((unsigned)(cblack[4] * cblack[5]))
+    cblack[6+c] -= i;
+  black += i;
+  if (env_black >= 0) black = env_black;
+  if (user_black >= 0) black = user_black;
+  FORC4 {
+    cblack[c] += black;
+    if (env_sat[c] > 0) {
+      max4[c] = env_sat[c];
+      if (maximum > env_sat[c]) maximum = env_sat[c];
+    }
+    if (user_sat[c] > 0) {
+      max4[c] = user_sat[c];
+      if (maximum > user_sat[c]) maximum = user_sat[c];
+    }
+  }
+}
+
+/*
    Identify which camera created this file, and set global variables
    accordingly.
  */
@@ -8946,7 +9060,7 @@ void CLASS identify()
     parse_fuji (get4());
     if (thumb_offset > 120) {
       fseek (ifp, 120, SEEK_SET);
-      is_raw += (i = get4()) && 1;
+      is_raw += (i = get4()) != 0;
       if (is_raw == 2 && shot_select)
 	parse_fuji (i);
     }
@@ -9155,6 +9269,7 @@ void CLASS identify()
       left_margin = 8;
       top_margin = 16;
     }
+    raw_color = 0;
   }
   for (i=0; i < sizeof unique / sizeof *unique; i++)
     if (unique_id == 0x80000000 + unique[i].id) {
@@ -9832,6 +9947,7 @@ dng_skip:
     if (raw_height < height) raw_height = height;
     if (raw_width  < width ) raw_width  = width;
   }
+  //dcraw_message (DCRAW_VERBOSE,_("raw color: %d\n"),raw_color);
   if (!tiff_bps) tiff_bps = 12;
   if (!maximum) maximum = (1 << tiff_bps) - 1;
   if (!load_raw || height < 22 || width < 22 ||
@@ -10332,7 +10448,8 @@ int CLASS main (int argc, const char **a
   static unsigned c;
   static int arg, status=0, quality, i;
   static int timestamp_only=0, thumbnail_only=0, identify_only=0;
-  static int user_qual=-1, user_black=-1, user_sat=-1, user_flip=-1;
+  static int user_qual=-1, user_black=-1, user_flip=-1;
+  static int user_sat[4] = { -1, -1, -1, -1 };
   static int use_fuji_rotate=1, write_to_stdout=0, read_from_stdin=0;
   static const char *sp, *bpfile=0, *dark_frame=0, *write_ext;
   static char opm, opt, *ofname, *cp;
@@ -10416,7 +10533,16 @@ int CLASS main (int argc, const char **a
 		 gamm[1] =     atof(argv[arg++]);
 		 if (gamm[0]) gamm[0] = 1/gamm[0]; break;
       case 'k':  user_black  = atoi(argv[arg++]);  break;
-      case 'S':  user_sat    = atoi(argv[arg++]);  break;
+      //case 'S':  user_sat    = atoi(argv[arg++]);  break;
+      case 'S':  {
+                 char const *e = argv[arg++];
+                 int v = sscanf(e, "%d%d%d%d", &user_sat[0], &user_sat[1], &user_sat[2], &user_sat[3]);
+                 if (1 == v) {
+                   user_sat[3] = user_sat[2] = user_sat[1] = user_sat[0];
+                 } else if (4 != v) {
+                   dcraw_message(DCRAW_WARNING, "cannot convert saturation value (%s) to integer(s)\n", e);
+                 }
+                 break; }
       case 't':  user_flip   = atoi(argv[arg++]);  break;
       case 'q':  user_qual   = atoi(argv[arg++]);  break;
       case 'm':  med_passes  = atoi(argv[arg++]);  break;
@@ -10462,6 +10588,7 @@ int CLASS main (int argc, const char **a
 	return 1;
     }
   }
+  read_env();
   if (arg == argc) {
     dcraw_message (DCRAW_ERROR,_("No files to process.\n"));
     return 1;
@@ -10661,19 +10788,7 @@ next:
     if (dark_frame) subtract (dark_frame);
     quality = 2 + !fuji_width;
     if (user_qual >= 0) quality = user_qual;
-    i = cblack[3];
-    FORC3 if (i > cblack[c]) i = cblack[c];
-    FORC4 cblack[c] -= i;
-    black += i;
-    i = cblack[6];
-    FORC ((unsigned)(cblack[4] * cblack[5]))
-      if (i > cblack[6+c]) i = cblack[6+c];
-    FORC ((unsigned)(cblack[4] * cblack[5]))
-      cblack[6+c] -= i;
-    black += i;
-    if (user_black >= 0) black = user_black;
-    FORC4 cblack[c] += black;
-    if (user_sat > 0) maximum = user_sat;
+    apply_black_sat(user_black, user_sat);
 #ifdef COLORCHECK
     colorcheck();
 #endif
